@@ -2,6 +2,11 @@ from abc import ABCMeta, abstractmethod
 import collections
 import sys
 
+try:
+    from collections import Counter
+except ImportError:
+    from counter import Counter
+
 
 class _Call(object):
     """ A genotype call, a cell entry in a VCF file"""
@@ -31,6 +36,8 @@ class _Call(object):
         """ Two _Calls are equal if their _Records are equal
             and the samples and ``gt_type``s are the same
         """
+        if not isinstance(other, self.__class__):
+            return False
         return (self.site == other.site
                 and self.sample == other.sample
                 and self.gt_type == other.gt_type)
@@ -143,27 +150,40 @@ class _Record(object):
         self.samples = samples or []
         self._sample_indexes = sample_indexes
 
+    # For Python 2
+    def __cmp__(self, other):
+        return cmp((self.CHROM, self.POS), (other.CHROM, other.POS))
+
+    # For Python 3
     def __eq__(self, other):
         """ _Records are equal if they describe the same variant (same position, alleles) """
+        # a _Record is never equal with a non-Record
+        # do the check here to avoid AttributeError (i.e. None does not have CHROM)
+        if not isinstance(other, self.__class__):
+            return False
         return (self.CHROM == other.CHROM and
                 self.POS == other.POS and
                 self.REF == other.REF and
                 self.ALT == other.ALT)
 
+    # For Python 3
+    def __lt__(self, other):
+        return (self.CHROM, self.POS) < (other.CHROM, other.POS)
+
     def __iter__(self):
         return iter(self.samples)
 
     def __str__(self):
         return "Record(CHROM=%(CHROM)s, POS=%(POS)s, REF=%(REF)s, ALT=%(ALT)s)" % self.__dict__
 
-    def __cmp__(self, other):
-        return cmp((self.CHROM, self.POS), (other.CHROM, other.POS))
-
     def add_format(self, fmt):
         self.FORMAT = self.FORMAT + ':' + fmt
 
     def add_filter(self, flt):
-        self.FILTER.append(flt)
+        if self.FILTER is None:
+            self.FILTER = [flt]
+        else:
+            self.FILTER.append(flt)
 
     def add_info(self, info, value=True):
         self.INFO[info] = value
@@ -204,17 +224,16 @@ class _Record(object):
 
     @property
     def aaf(self):
-        """ The allele frequency of the alternate allele.
-           NOTE 1: Punt if more than one alternate allele.
-           NOTE 2: Denominator calc'ed from _called_ genotypes.
+        """ A list of allele frequencies of alternate alleles.
+           NOTE: Denominator calc'ed from _called_ genotypes.
         """
-        # skip if more than one alternate allele. assumes bi-allelic
-        if len(self.ALT) > 1:
-            return None
-        het = self.num_het
-        hom_alt = self.num_hom_alt
-        num_chroms = float(2.0 * self.num_called)
-        return float(het + 2 * hom_alt) / float(num_chroms)
+        num_chroms = 2.0 * self.num_called
+        allele_counts = Counter()
+        for s in self.samples:
+            if s.gt_type is not None:
+                allele_counts.update([s.gt_alleles[0]])
+                allele_counts.update([s.gt_alleles[1]])
+        return [allele_counts[str(i)]/num_chroms for i in range(1, len(self.ALT)+1)]
 
     @property
     def nucl_diversity(self):
@@ -231,11 +250,23 @@ class _Record(object):
         # skip if more than one alternate allele. assumes bi-allelic
         if len(self.ALT) > 1:
             return None
-        p = self.aaf
+        p = self.aaf[0]
         q = 1.0 - p
         num_chroms = float(2.0 * self.num_called)
         return float(num_chroms / (num_chroms - 1.0)) * (2.0 * p * q)
 
+    @property
+    def heterozygosity(self):
+        """
+        Heterozygosity of a site. Heterozygosity gives the probability that
+        two randomly chosen chromosomes from the population have different
+        alleles, giving a measure of the degree of polymorphism in a population.
+
+        If there are i alleles with frequency p_i, H=1-sum_i(p_i^2)
+        """
+        allele_freqs = [1-sum(self.aaf)] + self.aaf
+        return 1 - sum(map(lambda x: x**2, allele_freqs))
+
     def get_hom_refs(self):
         """ The list of hom ref genotypes"""
         return [s for s in self.samples if s.gt_type == 0]
@@ -434,6 +465,8 @@ class _AltRecord(object):
         raise NotImplementedError
 
     def __eq__(self, other):
+        if not isinstance(other, self.__class__):
+            return False
         return self.type == other.type
 
 
@@ -460,8 +493,9 @@ class _Substitution(_AltRecord):
     def __eq__(self, other):
         if isinstance(other, basestring):
             return self.sequence == other
-        else:
-            return super(_Substitution, self).__eq__(other) and self.sequence == other.sequence
+        elif not isinstance(other, self.__class__):
+            return False
+        return super(_Substitution, self).__eq__(other) and self.sequence == other.sequence
 
 
 class _Breakend(_AltRecord):
@@ -470,9 +504,15 @@ class _Breakend(_AltRecord):
     def __init__(self, chr, pos, orientation, remoteOrientation, connectingSequence, withinMainAssembly, **kwargs):
         super(_Breakend, self).__init__(type="BND", **kwargs)
         #: The chromosome of breakend's mate.
-        self.chr = str(chr)
+        if chr is not None:
+            self.chr = str(chr)
+        else:
+            self.chr = None  # Single breakend
         #: The coordinate of breakend's mate.
-        self.pos = int(pos)
+        if pos is not None:
+            self.pos = int(pos)
+        else:
+            self.pos = None
         #: The orientation of breakend's mate. If the sequence 3' of the breakend's mate is connected, True, else if the sequence 5' of the breakend's mate is connected, False.
         self.remoteOrientation = remoteOrientation
         #: If the breakend mate is within the assembly, True, else False if the breakend mate is on a contig in an ancillary assembly file.
@@ -504,6 +544,8 @@ class _Breakend(_AltRecord):
             return self.connectingSequence + remoteTag
 
     def __eq__(self, other):
+        if not isinstance(other, self.__class__):
+            return False
         return super(_Breakend, self).__eq__(other) \
                 and self.chr == other.chr \
                 and self.pos == other.pos \
