@@ -236,6 +236,80 @@ class Test1kgSites(unittest.TestCase):
             assert not line.endswith('\t')
 
 
+class TestGoNL(unittest.TestCase):
+
+    def testParse(self):
+        reader = vcf.Reader(fh('gonl.chr20.release4.gtc.vcf'))
+        for _ in reader:
+            pass
+
+    def test_contig_line(self):
+        reader = vcf.Reader(fh('gonl.chr20.release4.gtc.vcf'))
+        self.assertEqual(reader.contigs['1'].length, 249250621)
+
+
+class TestInfoOrder(unittest.TestCase):
+
+    def _assert_order(self, definitions, fields):
+        """
+        Elements common to both lists should be in the same order. Elements
+        only in `fields` should be last and in alphabetical order.
+        """
+        used_definitions = [d for d in definitions if d in fields]
+        self.assertEqual(used_definitions, fields[:len(used_definitions)])
+        self.assertEqual(fields[len(used_definitions):],
+                         sorted(fields[len(used_definitions):]))
+
+    def test_writer(self):
+        """
+        Order of INFO fields should be compatible with the order of their
+        definition in the header and undefined fields should be last and in
+        alphabetical order.
+        """
+        reader = vcf.Reader(fh('1kg.sites.vcf', 'r'))
+        out = StringIO()
+        writer = vcf.Writer(out, reader, lineterminator='\n')
+
+        for record in reader:
+            writer.write_record(record)
+        out.seek(0)
+        out_str = out.getvalue()
+
+        definitions = []
+        for line in out_str.split('\n'):
+            if line.startswith('##INFO='):
+                definitions.append(line.split('ID=')[1].split(',')[0])
+            if not line or line.startswith('#'):
+                continue
+            fields = [f.split('=')[0] for f in line.split('\t')[7].split(';')]
+            self._assert_order(definitions, fields)
+
+
+class TestInfoTypeCharacter(unittest.TestCase):
+    def test_parse(self):
+        reader = vcf.Reader(fh('info-type-character.vcf'))
+        record = next(reader)
+        self.assertEqual(record.INFO['FLOAT_1'], 123.456)
+        self.assertEqual(record.INFO['CHAR_1'], 'Y')
+        self.assertEqual(record.INFO['FLOAT_N'], [123.456])
+        self.assertEqual(record.INFO['CHAR_N'], ['Y'])
+
+    def test_write(self):
+        reader = vcf.Reader(fh('info-type-character.vcf'))
+        out = StringIO()
+        writer = vcf.Writer(out, reader)
+
+        records = list(reader)
+
+        for record in records:
+            writer.write_record(record)
+        out.seek(0)
+        reader2 = vcf.Reader(out)
+
+        for l, r in zip(records, reader2):
+            self.assertEquals(l.INFO, r.INFO)
+
+
 class TestGatkOutputWriter(unittest.TestCase):
 
     def testWrite(self):
@@ -259,6 +333,7 @@ class TestGatkOutputWriter(unittest.TestCase):
         self.assertEquals(reader.samples, reader2.samples)
         self.assertEquals(reader.formats, reader2.formats)
         self.assertEquals(reader.infos, reader2.infos)
+        self.assertEquals(reader.contigs, reader2.contigs)
 
         for l, r in zip(records, reader2):
             self.assertEquals(l.samples, r.samples)
@@ -327,6 +402,20 @@ class TestSamplesSpace(unittest.TestCase):
         self.assertEqual(self.reader.samples, self.samples)
 
 
+class TestMixedFiltering(unittest.TestCase):
+    filename = 'mixed-filtering.vcf'
+    def test_mixed_filtering(self):
+        """
+        Test mix of FILTER values (pass, filtered, no filtering).
+        """
+        reader = vcf.Reader(fh(self.filename))
+        self.assertEqual(next(reader).FILTER, [])
+        self.assertEqual(next(reader).FILTER, ['q10'])
+        self.assertEqual(next(reader).FILTER, [])
+        self.assertEqual(next(reader).FILTER, None)
+        self.assertEqual(next(reader).FILTER, ['q10', 'q50'])
+
+
 class TestRecord(unittest.TestCase):
 
     def test_num_calls(self):
@@ -336,6 +425,11 @@ class TestRecord(unittest.TestCase):
                          var.num_het + var.num_unknown)
             self.assertEqual(len(var.samples), num_calls)
 
+    def test_dunder_eq(self):
+        rec = vcf.Reader(fh('example-4.0.vcf')).next()
+        self.assertFalse(rec == None)
+        self.assertFalse(None == rec)
+
     def test_call_rate(self):
         reader = vcf.Reader(fh('example-4.0.vcf'))
         for var in reader:
@@ -356,15 +450,15 @@ class TestRecord(unittest.TestCase):
         for var in reader:
             aaf = var.aaf
             if var.POS == 14370:
-                self.assertEqual(3.0/6.0, aaf)
+                self.assertEqual([3.0/6.0], aaf)
             if var.POS == 17330:
-                self.assertEqual(1.0/6.0, aaf)
+                self.assertEqual([1.0/6.0], aaf)
             if var.POS == 1110696:
-                self.assertEqual(None, aaf)
+                self.assertEqual([2.0/6.0, 4.0/6.0], aaf)
             if var.POS == 1230237:
-                self.assertEqual(0.0/6.0, aaf)
+                self.assertEqual([0.0/6.0], aaf)
             elif var.POS == 1234567:
-                self.assertEqual(None, aaf)
+                self.assertEqual([2.0/4.0, 1.0/4.0], aaf)
 
     def test_pi(self):
         reader = vcf.Reader(fh('example-4.0.vcf'))
@@ -381,6 +475,21 @@ class TestRecord(unittest.TestCase):
             elif var.POS == 1234567:
                 self.assertEqual(None, pi)
 
+    def test_heterozygosity(self):
+        reader = vcf.Reader(fh('example-4.0.vcf'))
+        for var in reader:
+            het = var.heterozygosity
+            if var.POS == 14370:
+                self.assertEqual(0.5, het)
+            if var.POS == 17330:
+                self.assertEqual(1-((1.0/6)**2 + (5.0/6)**2), het)
+            if var.POS == 1110696:
+                self.assertEqual(4.0/9.0, het)
+            if var.POS == 1230237:
+                self.assertEqual(0.0, het)
+            elif var.POS == 1234567:
+                self.assertEqual(5.0/8.0, het)
+
     def test_is_snp(self):
         reader = vcf.Reader(fh('example-4.0.vcf'))
         for r in reader:
@@ -644,6 +753,13 @@ class TestRecord(unittest.TestCase):
 
 class TestCall(unittest.TestCase):
 
+    def test_dunder_eq(self):
+        reader = vcf.Reader(fh('example-4.0.vcf'))
+        var = reader.next()
+        example_call = var.samples[0]
+        self.assertFalse(example_call == None)
+        self.assertFalse(None == example_call)
+
     def test_phased(self):
         reader = vcf.Reader(fh('example-4.0.vcf'))
         for var in reader:
@@ -861,6 +977,55 @@ class TestUtils(unittest.TestCase):
                 assert recs[0] is not None
                 assert recs[1] is not None
 
+        # case with working custom equality function
+
+        # without custom function, most records in these files
+        # are different since the default equality checks
+        # for ALT values
+
+        reader1 = vcf.Reader(fh('example-4.0.vcf'))
+        reader2 = vcf.Reader(fh('walk_refcall.vcf'))
+
+        # counters for distinct records and overlapping records
+        ndist_def, nover_def = 0, 0
+        for x in utils.walk_together(reader1, reader2):
+            assert len(x) == 2
+            if x[0] is not None and x[1] is not None:
+                assert (x[0] == x[1] and x[1] == x[0])
+                nover_def += 1
+            ndist_def += 1
+        # check how many overlapping records
+        assert nover_def == 1
+        # check how many distinct records
+        assert ndist_def == 8
+
+        # with custom function that does not check ALT,
+        # we see more overlaps and less distinct records
+
+        def custom_eq(rec1, rec2):
+            # check for equality only on CHROM, POS, and REF
+            if rec1 is None or rec2 is None:
+                return False
+            return rec1.CHROM == rec2.CHROM and rec1.POS == rec2.POS and \
+                    rec1.REF == rec2.REF
+
+        reader1 = vcf.Reader(fh('example-4.0.vcf'))
+        reader2 = vcf.Reader(fh('walk_refcall.vcf'))
+
+        ndist_cust, nover_cust = 0, 0
+        for x in utils.walk_together(reader1, reader2, eq_func=custom_eq):
+            assert len(x) == 2
+            if x[0] is not None and x[1] is not None:
+                assert (custom_eq(x[0], x[1]) and custom_eq(x[1], x[0]))
+                nover_cust += 1
+            ndist_cust += 1
+        assert nover_cust == 4
+        assert ndist_cust == 5
+
+        # final check just to be absolutely sure
+        assert ndist_def != ndist_cust
+        assert nover_def != nover_cust
+
     def test_trim(self):
         tests = [('TAA GAA', 'T G'),
                  ('TA TA', 'T T'),
@@ -884,21 +1049,26 @@ class TestGATKMeta(unittest.TestCase):
         assert reader.metadata['GATKCommandLine'][1]['CommandLineOptions'] == '"analysis_type=VariantAnnotator annotation=[HomopolymerRun, VariantType, TandemRepeatAnnotator]"'
 
 
-
+suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestVcfSpecs))
 suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestGatkOutput))
 suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestFreebayesOutput))
 suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestSamtoolsOutput))
 suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestBcfToolsOutput))
+suite.addTests(unittest.TestLoader().loadTestsFromTestCase(Test1kg))
+suite.addTests(unittest.TestLoader().loadTestsFromTestCase(Test1kgSites))
+suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestGoNL))
+suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestInfoOrder))
+suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestInfoTypeCharacter))
 suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestGatkOutputWriter))
 suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestBcfToolsOutputWriter))
 suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestWriterDictionaryMeta))
+suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestSamplesSpace))
+suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestMixedFiltering))
+suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestRecord))
+suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestCall))
 suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestTabix))
 suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestOpenMethods))
 suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestFilter))
-suite.addTests(unittest.TestLoader().loadTestsFromTestCase(Test1kg))
-suite.addTests(unittest.TestLoader().loadTestsFromTestCase(Test1kgSites))
-suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestSamplesSpace))
-suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestRecord))
-suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestCall))
 suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestRegression))
-suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestVcfSpecs))
+suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestUtils))
+suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestGATKMeta))
